"""Tests for runway.cfngin.context."""
# pylint: disable=protected-access,too-many-public-methods
import io
import json
import unittest

from botocore.exceptions import ClientError
from botocore.response import StreamingBody
from botocore.stub import ANY, Stubber
from mock import PropertyMock, patch

from runway.cfngin.config import Config, load
from runway.cfngin.context import Context, get_fqn
from runway.cfngin.exceptions import (PersistentGraphCannotLock,
                                      PersistentGraphCannotUnlock,
                                      PersistentGraphLockCodeMissmatch,
                                      PersistentGraphLocked,
                                      PersistentGraphUnlocked)
from runway.cfngin.hooks.utils import handle_hooks
from runway.cfngin.plan import Graph, json_serial

BOTO3_CREDENTIALS = {
    'aws_access_key_id': 'foo',
    'aws_secret_access_key': 'bar',
    'aws_session_token': 'foobar'
}
GET_SESSION_CALL = {
    'access_key': BOTO3_CREDENTIALS['aws_access_key_id'],
    'secret_key': BOTO3_CREDENTIALS['aws_secret_access_key'],
    'session_token': BOTO3_CREDENTIALS['aws_session_token']
}


def gen_tagset(tags):
    """Create TagSet value from a dict."""
    return [{'Key': key, 'Value': value} for key, value in tags.items()]


def gen_s3_object_content(content):
    """Convert a string or dict to S3 object body.

    Args:
        content (Union[str, Dict[str, Any]]): S3 object body

    Returns:
        botocore.response.StreamingBody Used in the Body of a
            s3.get_object response.

    """
    if isinstance(content, dict):
        content = json.dumps(content, default=json_serial)
    encoded_content = content.encode()
    return StreamingBody(io.BytesIO(encoded_content),
                         len(encoded_content))


class TestContext(unittest.TestCase):
    """Tests for runway.cfngin.context.Context."""

    def setUp(self):
        """Run before tests."""
        self.config = Config({
            "namespace": "namespace",
            "stacks": [
                {"name": "stack1"}, {"name": "stack2"}]})
        self.persist_graph_raw_config = {
            'namespace': 'test',
            'cfngin_bucket': 'cfngin-test',
            'cfngin_bucket_region': 'us-east-1',
            'persistent_graph_key': 'test.json',
            'stacks': [
                {'name': 'stack1'}, {'name': 'stack2', 'requires': ['stack1']}]
        }
        self.persist_graph_config = Config(self.persist_graph_raw_config)

    def test_context_optional_keys_set(self):
        """Test context optional keys set."""
        context = Context(
            config=Config({}),
            stack_names=["stack"],
        )
        self.assertEqual(context.mappings, {})
        self.assertEqual(context.stack_names, ["stack"])

    def test_context_get_stacks(self):
        """Test context get stacks."""
        context = Context(config=self.config)
        self.assertEqual(len(context.get_stacks()), 2)

    def test_context_get_stacks_dict_use_fqn(self):
        """Test context get stacks dict use fqn."""
        context = Context(config=self.config)
        stacks_dict = context.get_stacks_dict()
        stack_names = sorted(stacks_dict.keys())
        self.assertEqual(stack_names[0], "namespace-stack1")
        self.assertEqual(stack_names[1], "namespace-stack2")

    def test_context_get_fqn(self):
        """Test context get fqn."""
        context = Context(config=self.config)
        fqn = context.get_fqn()
        self.assertEqual(fqn, "namespace")

    def test_context_get_fqn_replace_dot(self):
        """Test context get fqn replace dot."""
        context = Context(config=Config({"namespace": "my.namespace"}))
        fqn = context.get_fqn()
        self.assertEqual(fqn, "my-namespace")

    def test_context_get_fqn_empty_namespace(self):
        """Test context get fqn empty namespace."""
        context = Context(config=Config({"namespace": ""}))
        fqn = context.get_fqn("vpc")
        self.assertEqual(fqn, "vpc")
        self.assertEqual(context.tags, {})

    def test_context_namespace(self):
        """Test context namespace."""
        context = Context(config=Config({"namespace": "namespace"}))
        self.assertEqual(context.namespace, "namespace")

    def test_context_get_fqn_stack_name(self):
        """Test context get fqn stack name."""
        context = Context(config=self.config)
        fqn = context.get_fqn("stack1")
        self.assertEqual(fqn, "namespace-stack1")

    def test_context_default_bucket_name(self):
        """Test context default bucket name."""
        context = Context(config=Config({"namespace": "test"}))
        self.assertEqual(context.bucket_name, "stacker-test")

    def test_context_bucket_name_is_overridden_but_is_none(self):
        """Test context bucket name is overridden but is none."""
        config = Config({"namespace": "test", "cfngin_bucket": ""})
        context = Context(config=config)
        self.assertEqual(context.bucket_name, None)

        config = Config({"namespace": "test", "cfngin_bucket": None})
        context = Context(config=config)
        self.assertEqual(context.bucket_name, "stacker-test")

    def test_context_bucket_name_is_overridden(self):
        """Test context bucket name is overridden."""
        config = Config({"namespace": "test", "cfngin_bucket": "bucket123"})
        context = Context(config=config)
        self.assertEqual(context.bucket_name, "bucket123")

    def test_context_default_bucket_no_namespace(self):
        """Test context default bucket no namespace."""
        context = Context(config=Config({"namespace": ""}))
        self.assertEqual(context.bucket_name, None)

        context = Context(config=Config({"namespace": None}))
        self.assertEqual(context.bucket_name, None)

        context = Context(
            config=Config({"namespace": None, "cfngin_bucket": ""}))
        self.assertEqual(context.bucket_name, None)

    def test_context_namespace_delimiter_is_overridden_and_not_none(self):
        """Test context namespace delimiter is overridden and not none."""
        config = Config({"namespace": "namespace", "namespace_delimiter": "_"})
        context = Context(config=config)
        fqn = context.get_fqn("stack1")
        self.assertEqual(fqn, "namespace_stack1")

    def test_context_namespace_delimiter_is_overridden_and_is_empty(self):
        """Test context namespace delimiter is overridden and is empty."""
        config = Config({"namespace": "namespace", "namespace_delimiter": ""})
        context = Context(config=config)
        fqn = context.get_fqn("stack1")
        self.assertEqual(fqn, "namespacestack1")

    def test_context_tags_with_empty_map(self):
        """Test context tags with empty map."""
        config = Config({"namespace": "test", "tags": {}})
        context = Context(config=config)
        self.assertEqual(context.tags, {})

    def test_context_no_tags_specified(self):
        """Test context no tags specified."""
        config = Config({"namespace": "test"})
        context = Context(config=config)
        self.assertEqual(context.tags, {"cfngin_namespace": "test"})

    @patch('runway.cfngin.context.get_session')
    def test_get_session(self, mock_get_session):
        """Test get_session."""
        creds = BOTO3_CREDENTIALS.copy()
        context = Context(boto3_credentials=creds,
                          region='us-east-1')

        context.get_session()
        mock_get_session.assert_called_with(region='us-east-1',
                                            **GET_SESSION_CALL)

        context.get_session(region='us-west-2')
        mock_get_session.assert_called_with(region='us-west-2',
                                            **GET_SESSION_CALL)

        context.get_session(profile='user')
        mock_get_session.assert_called_with(region='us-east-1',
                                            profile='user')

    def test_hook_with_sys_path(self):
        """Test hook with sys path."""
        config = Config({
            "namespace": "test",
            "sys_path": "./tests/cfngin",
            "pre_build": [
                {
                    "data_key": "myHook",
                    "path": "fixtures.mock_hooks.mock_hook",
                    "required": True,
                    "args": {
                        "value": "mockResult"}}]})
        load(config)
        context = Context(config=config)
        stage = "pre_build"
        handle_hooks(stage, context.config[stage], "mock-region-1", context)
        self.assertEqual("mockResult", context.hook_data["myHook"]["result"])

    def test_persistent_graph_location(self):
        """Test persistent graph location."""
        context = Context(config=self.persist_graph_config)
        expected = {
            'Bucket': 'cfngin-test',
            'Key': 'persistent_graphs/test/test.json'
        }
        self.assertEqual(expected, context.persistent_graph_location)

    def test_persistent_graph_location_no_json(self):
        """'.json' appended to the key if it does not exist."""
        cp_config = self.persist_graph_raw_config.copy()
        cp_config['persistent_graph_key'] = 'test'

        context = Context(config=Config(cp_config))
        expected = {
            'Bucket': 'cfngin-test',
            'Key': 'persistent_graphs/test/test.json'
        }
        self.assertEqual(expected, context.persistent_graph_location)

    def test_persistent_graph_location_no_key(self):
        """Return an empty dict if key is not set."""
        context = Context(config=self.config)
        self.assertEqual({}, context.persistent_graph_location)

    def test_persistent_graph_location_no_bucket(self):
        """Return an empty dict if key is set but no bucket name."""
        cp_config = self.persist_graph_raw_config.copy()
        cp_config['cfngin_bucket'] = ''

        context = Context(config=Config(cp_config))
        self.assertEqual({}, context.persistent_graph_location)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    def test_persistent_graph_lock_code_disabled(self, mock_prop):
        """Return 'None' when not used."""
        mock_prop.return_value = None
        context = Context(config=Config(self.config))
        mock_prop.assert_not_called()
        self.assertIsNone(context.persistent_graph_lock_code)

    def test_persistent_graph_lock_code_present(self):
        """Return the value of the lock tag when it exists."""
        context = Context(config=self.persist_graph_config)
        stubber = Stubber(context.s3_conn)
        code = '0000'

        stubber.add_response('get_object_tagging', {
            'TagSet': gen_tagset({context._persistent_graph_lock_tag: code})
        }, context.persistent_graph_location)

        with stubber:
            self.assertIsNone(context._persistent_graph_lock_code)
            self.assertEqual(code, context.persistent_graph_lock_code)
            self.assertEqual(code, context._persistent_graph_lock_code)
            stubber.assert_no_pending_responses()

    def test_persistent_graph_lock_code_none(self):
        """Return 'None' when the tag is not set."""
        context = Context(config=self.persist_graph_config)
        stubber = Stubber(context.s3_conn)

        stubber.add_response('get_object_tagging', {'TagSet': []},
                             context.persistent_graph_location)

        with stubber:
            self.assertIsNone(context.persistent_graph_lock_code)
            self.assertIsNone(context._persistent_graph_lock_code)
            stubber.assert_no_pending_responses()

    def test_persistent_graph_lock_code_no_object(self):
        """Return 'None' when object does not exist."""
        context = Context(config=self.persist_graph_config)
        stubber = Stubber(context.s3_conn)

        stubber.add_client_error(
            'get_object_tagging', 'NoSuchKey',
            expected_params=context.persistent_graph_location)

        with stubber:
            self.assertIsNone(context.persistent_graph_lock_code)
            self.assertIsNone(context._persistent_graph_lock_code)
            stubber.assert_no_pending_responses()

    def test_persistent_graph(self):
        """Return Graph from S3 object."""
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        stubber = Stubber(context.s3_conn)
        expected_params = {'ResponseContentType': 'application/json'}
        expected_params.update(context.persistent_graph_location)
        expected_content = {
            'stack1': set(),
            'stack2': set(['stack1'])
        }

        stubber.add_response('get_object',
                             {'Body': gen_s3_object_content(expected_content)},
                             expected_params)

        with stubber:
            self.assertIsNone(context._persistent_graph)
            self.assertIsInstance(context.persistent_graph, Graph)
            self.assertIsInstance(context._persistent_graph, Graph)
            self.assertEqual(expected_content,
                             context.persistent_graph.to_dict())
            stubber.assert_no_pending_responses()

    def test_persistent_graph_no_object(self):
        """Create object if one does not exist and return empty Graph."""
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        stubber = Stubber(context.s3_conn)
        expected_get_params = {'ResponseContentType': 'application/json'}
        expected_get_params.update(context.persistent_graph_location)
        expected_put_params = {'Body': '{}',
                               'ServerSideEncryption': 'AES256',
                               'ACL': 'bucket-owner-full-control',
                               'ContentType': 'application/json'}
        expected_put_params.update(context.persistent_graph_location)

        stubber.add_client_error('get_object', 'NoSuchKey',
                                 expected_params=expected_get_params)
        stubber.add_response('put_object', {}, expected_put_params)

        with stubber:
            self.assertIsNone(context._persistent_graph)
            self.assertIsInstance(context.persistent_graph, Graph)
            self.assertIsInstance(context._persistent_graph, Graph)
            self.assertEqual({}, context.persistent_graph.to_dict())
            stubber.assert_no_pending_responses()

    def test_persistent_graph_disabled(self):
        """Return 'None' when key is not set."""
        context = Context(config=self.config)
        self.assertIsNone(context._persistent_graph)
        self.assertIsNone(context.persistent_graph)

    def test_lock_persistent_graph(self):
        """Return 'None' when lock is successful."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph()
        stubber = Stubber(context.s3_conn)
        expected_params = {'Tagging': {
            'TagSet': gen_tagset({context._persistent_graph_lock_tag: code})
        }}
        expected_params.update(context.persistent_graph_location)

        stubber.add_response('get_object_tagging', {'TagSet': []},
                             context.persistent_graph_location)
        stubber.add_response('put_object_tagging', {}, expected_params)

        with stubber:
            self.assertIsNone(context.lock_persistent_graph(code))
            stubber.assert_no_pending_responses()

    def test_lock_persistent_graph_locked(self):
        """Error raised when when object is locked."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph()
        stubber = Stubber(context.s3_conn)
        expected_params = {'Tagging': {
            'TagSet': gen_tagset({context._persistent_graph_lock_tag: code})
        }}
        expected_params.update(context.persistent_graph_location)

        stubber.add_response('get_object_tagging',
                             {'TagSet': gen_tagset(
                                 {context._persistent_graph_lock_tag: '1111'}
                             )},
                             context.persistent_graph_location)

        with stubber:
            with self.assertRaises(PersistentGraphLocked):
                context.lock_persistent_graph(code)
            stubber.assert_no_pending_responses()

    def test_lock_persistent_graph_no_object(self):
        """Error raised when when there is no object to lock."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph()
        stubber = Stubber(context.s3_conn)
        expected_params = {'Tagging': {
            'TagSet': gen_tagset({context._persistent_graph_lock_tag: code})
        }}
        expected_params.update(context.persistent_graph_location)

        stubber.add_client_error(
            'get_object_tagging', 'NoSuchKey',
            expected_params=context.persistent_graph_location
        )
        stubber.add_client_error('put_object_tagging', 'NoSuchKey',
                                 expected_params=expected_params)

        with stubber:
            with self.assertRaises(PersistentGraphCannotLock):
                context.lock_persistent_graph(code)
            stubber.assert_no_pending_responses()

    def test_put_persistent_graph(self):
        """Return 'None' when put is successful."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        graph_dict = {
            'stack1': [],
            'stack2': ['stack1']
        }
        context._persistent_graph = Graph.from_dict(graph_dict, context)
        stubber = Stubber(context.s3_conn)
        expected_params = {'Body': json.dumps(graph_dict, indent=4),
                           'ServerSideEncryption': 'AES256',
                           'ACL': 'bucket-owner-full-control',
                           'ContentType': 'application/json',
                           'Tagging': '{}={}'.format(
                               context._persistent_graph_lock_tag,
                               code)}
        expected_params.update(context.persistent_graph_location)

        stubber.add_response('get_object_tagging',
                             {'TagSet': gen_tagset(
                                 {context._persistent_graph_lock_tag: code}
                             )},
                             context.persistent_graph_location)
        stubber.add_response('put_object', {}, expected_params)

        with stubber:
            self.assertIsNone(context.put_persistent_graph(code))
            stubber.assert_no_pending_responses()

    def test_put_persistent_graph_unlocked(self):
        """Error raised when trying to update an unlocked object."""
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph.from_dict({'stack1': []}, context)
        stubber = Stubber(context.s3_conn)

        stubber.add_response('get_object_tagging', {'TagSet': []},
                             context.persistent_graph_location)

        with stubber:
            with self.assertRaises(PersistentGraphUnlocked):
                context.put_persistent_graph('')
            stubber.assert_no_pending_responses()

    def test_put_persistent_graph_code_missmatch(self):
        """Error raised when provided lock code does not match object."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph.from_dict({'stack1': []}, context)
        stubber = Stubber(context.s3_conn)

        stubber.add_response('get_object_tagging',
                             {'TagSet': gen_tagset(
                                 {context._persistent_graph_lock_tag: '1111'}
                             )},
                             context.persistent_graph_location)

        with stubber:
            with self.assertRaises(PersistentGraphLockCodeMissmatch):
                context.put_persistent_graph(code)
            stubber.assert_no_pending_responses()

    def test_put_persistent_graph_empty(self):
        """Object deleted when persistent graph is empty."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph()
        stubber = Stubber(context.s3_conn)

        stubber.add_response('delete_object', {},
                             context.persistent_graph_location)

        with stubber:
            self.assertFalse(context.persistent_graph.to_dict())
            self.assertIsNone(context.put_persistent_graph(code))
            stubber.assert_no_pending_responses()

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    def test_persistent_graph_locked(self, mock_prop):
        """Return 'True' or 'False' based on code property."""
        mock_prop.return_value = {}
        context = Context(config=self.persist_graph_config)
        context._persistent_graph = True

        context._persistent_graph_lock_code = True
        self.assertTrue(context.persistent_graph_locked)

        context._persistent_graph_lock_code = None
        self.assertFalse(context.persistent_graph_locked)
        mock_prop.assert_called_once()

    def test_persistent_graph_locked_disabled(self):
        """Return 'None' when key is not set."""
        context = Context(config=self.config)
        self.assertFalse(context.persistent_graph_locked)

    def test_s3_bucket_exists(self):
        """Test s3 bucket exists."""
        context = Context(config=self.config)
        stubber = Stubber(context.s3_conn)

        stubber.add_response(
            "head_bucket",
            service_response={},
            expected_params={
                "Bucket": ANY,
            }
        )

        with stubber:
            self.assertIsNone(context._s3_bucket_verified)
            self.assertTrue(context.s3_bucket_verified)
            self.assertTrue(context._s3_bucket_verified)
            stubber.assert_no_pending_responses()

    def test_s3_bucket_does_not_exist_us_east(self):
        """Create S3 bucket when it does not exist."""
        context = Context(config=self.config, region='us-east-1')
        stubber = Stubber(context.s3_conn)

        stubber.add_client_error(
            "head_bucket",
            service_error_code="NoSuchBucket",
            service_message="Not Found",
            http_status_code=404,
        )
        stubber.add_response(
            "create_bucket",
            service_response={},
            expected_params={
                "Bucket": ANY,
            }
        )

        with stubber:
            self.assertIsNone(context._s3_bucket_verified)
            self.assertTrue(context.s3_bucket_verified)
            self.assertTrue(context._s3_bucket_verified)
            stubber.assert_no_pending_responses()

    def test_s3_bucket_does_not_exist_us_west(self):
        """Create S3 bucket with loc constraints when it does not exist."""
        region = 'us-west-1'
        context = Context(config=self.config, region=region)
        stubber = Stubber(context.s3_conn)

        stubber.add_client_error(
            "head_bucket",
            service_error_code="NoSuchBucket",
            service_message="Not Found",
            http_status_code=404,
        )
        stubber.add_response(
            "create_bucket",
            service_response={},
            expected_params={
                "Bucket": ANY,
                "CreateBucketConfiguration": {
                    "LocationConstraint": region,
                }
            }
        )

        with stubber:
            self.assertIsNone(context._s3_bucket_verified)
            self.assertTrue(context.s3_bucket_verified)
            self.assertTrue(context._s3_bucket_verified)
            stubber.assert_no_pending_responses()

    def test_s3_bucket_forbidden(self):
        """Error raised when S3 bucket exists but cannot access."""
        context = Context(config=self.config)
        stubber = Stubber(context.s3_conn)

        stubber.add_client_error(
            "head_bucket",
            service_error_code="AccessDenied",
            service_message="Forbidden",
            http_status_code=403,
        )

        with stubber:
            with self.assertRaises(ClientError):
                self.assertFalse(context.s3_bucket_verified)
            stubber.assert_no_pending_responses()

    def test_unlock_persistent_graph(self):
        """Return 'True' when delete tag is successful."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph.from_dict({'stack1': []}, context)
        stubber = Stubber(context.s3_conn)

        stubber.add_response('get_object_tagging',
                             {'TagSet': gen_tagset(
                                 {context._persistent_graph_lock_tag: code}
                             )},
                             context.persistent_graph_location)
        stubber.add_response('delete_object_tagging', {},
                             context.persistent_graph_location)

        with stubber:
            self.assertTrue(context.unlock_persistent_graph(code))
            stubber.assert_no_pending_responses()

    def test_unlock_persistent_graph_not_locked(self):
        """Error raised when object is not locked."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph.from_dict({'stack1': []}, context)
        stubber = Stubber(context.s3_conn)

        stubber.add_response('get_object_tagging',
                             {'TagSet': []},
                             context.persistent_graph_location)

        with stubber:
            with self.assertRaises(PersistentGraphCannotUnlock):
                context.unlock_persistent_graph(code)
            stubber.assert_no_pending_responses()

    def test_unlock_persistent_graph_code_missmatch(self):
        """Error raised when local code does not match object."""
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph.from_dict({'stack1': []}, context)
        stubber = Stubber(context.s3_conn)

        stubber.add_response('get_object_tagging',
                             {'TagSet': gen_tagset(
                                 {context._persistent_graph_lock_tag: '1111'}
                             )},
                             context.persistent_graph_location)

        with stubber:
            with self.assertRaises(PersistentGraphCannotUnlock):
                context.unlock_persistent_graph(code)
            stubber.assert_no_pending_responses()

    def test_unlock_persistent_graph_no_object(self):
        """Return 'None' when object does not exist.

        This can occur if the object is deleted by 'put_persistent_graph'.

        """
        code = '0000'
        context = Context(config=self.persist_graph_config)
        context._s3_bucket_verified = True
        context._persistent_graph = Graph()
        stubber = Stubber(context.s3_conn)
        expected_params = context.persistent_graph_location.copy()
        expected_params.update({'ResponseContentType': 'application/json'})

        stubber.add_client_error(
            'get_object', 'NoSuchKey',
            expected_params=expected_params
        )

        with stubber:
            assert context.unlock_persistent_graph(code)
            stubber.assert_no_pending_responses()


class TestFunctions(unittest.TestCase):
    """Test the module level functions."""

    def test_get_fqn_redundant_base(self):
        """Test get fqn redundant base."""
        base = "woot"
        name = "woot-blah"
        self.assertEqual(get_fqn(base, '-', name), name)
        self.assertEqual(get_fqn(base, '', name), name)
        self.assertEqual(get_fqn(base, '_', name), "woot_woot-blah")

    def test_get_fqn_only_base(self):
        """Test get fqn only base."""
        base = "woot"
        self.assertEqual(get_fqn(base, '-'), base)
        self.assertEqual(get_fqn(base, ''), base)
        self.assertEqual(get_fqn(base, '_'), base)

    def test_get_fqn_full(self):
        """Test get fqn full."""
        base = "woot"
        name = "blah"
        self.assertEqual(get_fqn(base, '-', name), "%s-%s" % (base, name))
        self.assertEqual(get_fqn(base, '', name), "%s%s" % (base, name))
        self.assertEqual(get_fqn(base, '_', name), "%s_%s" % (base, name))


if __name__ == '__main__':
    unittest.main()



"""Tests for runway.cfngin.blueprints.testutil."""
import unittest

from troposphere import ecr

from runway.cfngin.blueprints.base import Blueprint
from runway.cfngin.blueprints.testutil import BlueprintTestCase
from runway.cfngin.context import Context
from runway.variables import Variable


class Repositories(Blueprint):
    """Simple blueprint to test our test cases."""

    VARIABLES = {
        "Repositories": {
            "type": list,
            "description": "A list of repository names to create."
        }
    }

    def create_template(self):
        """Create template."""
        template = self.template
        variables = self.get_variables()

        for repo in variables["Repositories"]:
            template.add_resource(
                ecr.Repository(
                    "%sRepository" % repo,
                    RepositoryName=repo,
                )
            )


class TestRepositories(BlueprintTestCase):
    """Tests for runway.cfngin.blueprints.testutil.BlueprintTestCase."""

    def test_create_template_passes(self):
        """Test create template passes."""
        ctx = Context({'namespace': 'test'})
        blueprint = Repositories('test_repo', ctx)
        blueprint.resolve_variables([
            Variable('Repositories', ["repo1", "repo2"], 'cfngin')
        ])
        blueprint.create_template()
        self.assertRenderedBlueprint(blueprint)

    def test_create_template_fails(self):
        """Test create template fails."""
        ctx = Context({'namespace': 'test'})
        blueprint = Repositories('test_repo', ctx)
        blueprint.resolve_variables([
            Variable('Repositories', ["repo1", "repo2", "repo3"], 'cfngin')
        ])
        blueprint.create_template()
        with self.assertRaises(AssertionError):
            self.assertRenderedBlueprint(blueprint)


if __name__ == '__main__':
    unittest.main()



"""Tests for runway.cfngin.lookups.handlers.hook_data."""
# pylint: disable=no-self-use
import pytest
from troposphere.awslambda import Code

from runway.cfngin.exceptions import FailedVariableLookup
from runway.variables import Variable


class TestHookDataLookup(object):
    """Tests for runway.cfngin.lookups.handlers.hook_data.HookDataLookup."""

    def test_handle(self, cfngin_context):
        """Test handle with simple usage."""
        cfngin_context.set_hook_data('fake_hook',
                                     {'nested': {'result': 'good'}})
        var_top = Variable('test', '${hook_data fake_hook}',
                           variable_type='cfngin')
        var_nested = Variable('test', '${hook_data fake_hook.nested.result}',
                              variable_type='cfngin')
        var_top.resolve(cfngin_context)
        var_nested.resolve(cfngin_context)

        assert var_top.value == {'nested': {'result': 'good'}}
        assert var_nested.value == 'good'

    def test_default(self, cfngin_context):
        """Test handle with a default value."""
        cfngin_context.set_hook_data('fake_hook',
                                     {'nested': {'result': 'good'}})
        var_top = Variable('test', '${hook_data bad_hook::default=something}',
                           variable_type='cfngin')
        var_nested = Variable('test',
                              '${hook_data fake_hook.bad.'
                              'result::default=something,load=json,get=key}',
                              variable_type='cfngin')
        var_top.resolve(cfngin_context)
        var_nested.resolve(cfngin_context)

        assert var_top.value == 'something'
        assert var_nested.value == 'something'

    def test_not_found(self, cfngin_context):
        """Test value not found and no default."""
        variable = Variable('test',
                            '${hook_data fake_hook.bad.result}',
                            variable_type='cfngin')
        with pytest.raises(FailedVariableLookup) as err:
            variable.resolve(cfngin_context)

        assert 'ValueError' in str(err.value)
        assert 'Could not find a value for "fake_hook.bad.result"' in str(err.value)

    def test_troposphere(self, cfngin_context):
        """Test with troposphere object like returned from lambda hook."""
        bucket = 'test-bucket'
        s3_key = 'lambda_functions/my_function'
        cfngin_context.set_hook_data('lambda',
                                     {'my_function': Code(
                                         S3Bucket=bucket,
                                         S3Key=s3_key
                                     )})
        var_bucket = Variable('test',
                              '${hook_data lambda.my_function::'
                              'load=troposphere,get=S3Bucket}',
                              variable_type='cfngin')
        var_key = Variable('test',
                           '${hook_data lambda.my_function::get=S3Key}',
                           variable_type='cfngin')
        var_bucket.resolve(cfngin_context)
        var_key.resolve(cfngin_context)

        assert var_bucket.value == bucket
        assert var_key.value == s3_key

    def test_legacy_valid_hook_data(self, cfngin_context):
        """Test valid hook data."""
        cfngin_context.set_hook_data('fake_hook', {'result': 'good'})
        variable = Variable('test', "${hook_data fake_hook::result}",
                            variable_type='cfngin')
        with pytest.warns(DeprecationWarning):
            variable.resolve(cfngin_context)
        assert variable.value == 'good'

    def test_legacy_invalid_hook_data(self, cfngin_context):
        """Test invalid hook data."""
        cfngin_context.set_hook_data('fake_hook', {'result': 'good'})
        variable = Variable('test', "${hook_data fake_hook::bad_key}",
                            variable_type='cfngin')
        with pytest.raises(FailedVariableLookup) as err, \
                pytest.warns(DeprecationWarning):
            variable.resolve(cfngin_context)

        assert 'ValueError' in str(err.value)

    def test_legacy_bad_value_hook_data(self, cfngin_context):
        """Test bad value hook data."""
        variable = Variable('test', "${hook_data fake_hook::bad_key}",
                            variable_type='cfngin')

        with pytest.raises(FailedVariableLookup) as err, \
                pytest.warns(DeprecationWarning):
            variable.resolve(cfngin_context)

        assert 'ValueError' in str(err.value)



"""CFNgin stack."""
from copy import deepcopy

from runway.util import load_object_from_string
from runway.variables import Variable, resolve_variables

from .blueprints.raw import RawTemplateBlueprint


def _initialize_variables(stack_def, variables=None):
    """Convert defined variables into a list of ``Variable`` for consumption.

    Args:
        stack_def (Dict[str, Any]): The stack definition being worked on.
        variables (Dict[str, Any]): Optional, explicit variables.

    Returns:
        List[Variable]: Contains key/value pairs of the collected variables.

    Raises:
        AttributeError: Raised when the stack definition contains an invalid
            attribute. Currently only when using old parameters, rather than
            variables.

    """
    variables = variables or stack_def.variables or {}
    variable_values = deepcopy(variables)
    return [Variable(k, v, 'cfngin') for k, v in variable_values.items()]


class Stack(object):
    """Represents gathered information about a stack to be built/updated.

    Attributes:
        definition (:class:`runway.cfngin.config.Stack`): The stack definition
            from the config.
        enabled (bool): Whether this stack is enabled
        force (bool): Whether to force updates on this stack.
        fqn (str): Fully qualified name of the stack. Combines the stack name
            and current namespace.
        in_progress_behavior (Optional[str]): The behavior for when a stack is
            in ``CREATE_IN_PROGRESS`` or ``UPDATE_IN_PROGRESS``.
        locked (bool): Whether or not the stack is locked.
        logging (bool): Whether logging is enabled.
        mappings (Optional[Dict[str, Dict[str, Any]]]): Cloudformation
            mappings passed to the blueprint.
        name (str): Name of the stack taken from the definition.
        outputs (Optional[Dict[str, Any]]): CloudFormation Stack outputs
        profile (str): Profile name from the stack definition.
        protected (bool): Whether this stack is protected.
        region (str): AWS region name.
        termination_protection (bool): The state of termination protection
            to apply to the stack.
        variables (Optional[Dict[str, Any]]): Variables for the stack.

    """

    def __init__(self, definition, context, variables=None, mappings=None,
                 locked=False, force=False, enabled=True, protected=False):
        """Instantiate class.

        Args:
            definition (:class:`runway.cfngin.config.Stack`): A stack
                definition.
            context (:class:`runway.cfngin.context.Context`): Current context
                for building the stack.
            variables (Optional[Dict[str, Any]]): Variables for the stack.
            mappings (Optional[Dict[str, Dict[str, Any]]]): Cloudformation
                mappings passed to the blueprint.
            locked (bool): Whether or not the stack is locked.
            force (bool): Whether to force updates on this stack.
            enabled (bool): Whether this stack is enabled
            protected (bool): Whether this stack is protected.

        """
        self._blueprint = None
        self._stack_policy = None

        self.name = definition.name  # dependency of other attrs
        self.context = context
        self.definition = definition
        self.enabled = enabled
        self.force = force
        self.fqn = context.get_fqn(definition.stack_name or self.name)
        self.in_progress_behavior = definition.in_progress_behavior
        self.locked = locked
        self.logging = True
        self.mappings = mappings
        self.outputs = None
        self.profile = definition.profile
        self.protected = protected
        self.region = definition.region
        self.termination_protection = definition.termination_protection
        self.variables = _initialize_variables(definition, variables)

    @property
    def required_by(self):
        """Return a list of stack names that depend on this stack.

        Returns:
            List[str]

        """
        return self.definition.required_by or []

    @property
    def requires(self):
        """Return a list of stack names this stack depends on.

        Returns:
            List[str]

        """
        requires = set(self.definition.requires or [])

        # Add any dependencies based on output lookups
        for variable in self.variables:
            deps = variable.dependencies
            if self.name in deps:
                message = (
                    "Variable %s in stack %s has a circular reference"
                ) % (variable.name, self.name)
                raise ValueError(message)
            requires.update(deps)
        return requires

    @property
    def stack_policy(self):
        """Return the Stack Policy to use for this stack."""
        if not self._stack_policy:
            self._stack_policy = None
            if self.definition.stack_policy_path:
                with open(self.definition.stack_policy_path) as file_:
                    self._stack_policy = file_.read()

        return self._stack_policy

    @property
    def blueprint(self):
        """Return the blueprint associated with this stack."""
        if not self._blueprint:
            kwargs = {}
            blueprint_class = None
            if self.definition.class_path:
                class_path = self.definition.class_path
                blueprint_class = load_object_from_string(class_path)
                if not hasattr(blueprint_class, "rendered"):
                    raise AttributeError("Stack class %s does not have a "
                                         "\"rendered\" "
                                         "attribute." % (class_path,))
            elif self.definition.template_path:
                blueprint_class = RawTemplateBlueprint
                kwargs["raw_template_path"] = self.definition.template_path
            else:
                raise AttributeError("Stack does not have a defined class or "
                                     "template path.")

            self._blueprint = blueprint_class(
                name=self.name,
                context=self.context,
                mappings=self.mappings,
                description=self.definition.description,
                **kwargs
            )
        return self._blueprint

    @property
    def tags(self):
        """Return the tags that should be set on this stack.

        Includes both the global tags, as well as any stack specific tags
        or overrides.

        Returns:
            Dict[str, str]: Dictionary of tags.

        """
        tags = self.definition.tags or {}
        return dict(self.context.tags, **tags)

    @property
    def parameter_values(self):
        """Return all CloudFormation Parameters for the stack.

        CloudFormation Parameters can be specified via Blueprint Variables
        with a :class:`runway.cfngin.blueprints.variables.types.CFNType`
        ``type``.

        Returns:
            Dict[str, Any]: dictionary of
            ``<parameter name>: <parameter value>``.

        """
        return self.blueprint.get_parameter_values()

    @property
    def all_parameter_definitions(self):
        """Return a list of all parameters in the blueprint/template.

        Dict[str, Dict[str, str]]: parameter definitions. Keys are
        parameter names, the values are dicts containing key/values
        for various parameter properties.

        """
        return self.blueprint.get_parameter_definitions()

    @property
    def required_parameter_definitions(self):
        """Return all CloudFormation Parameters without a default value.

        Returns:
            Dict[str, Dict[str, str]]: dict of required CloudFormation
            Parameters for the blueprint. Will be a dictionary of
            ``<parameter name>: <parameter attributes>``.

        """
        return self.blueprint.get_required_parameter_definitions()

    def resolve(self, context, provider):
        """Resolve the Stack variables.

        This resolves the Stack variables and then prepares the Blueprint for
        rendering by passing the resolved variables to the Blueprint.

        Args:
            context (:class:`runway.cfngin.context.Context`): CFNgin context.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Subclass of the base provider.

        """
        resolve_variables(self.variables, context, provider)
        self.blueprint.resolve_variables(self.variables)

    def set_outputs(self, outputs):
        """Set stack outputs to the provided value.

        Args:
            outputs (Dict[str, Any]): CloudFormation Stack outputs.

        """
        self.outputs = outputs

    def __repr__(self):
        """Object represented as a string."""
        return self.fqn



"""Tests for runway.cfngin.lookups.handlers.ami."""
import unittest

import boto3
import mock
from botocore.stub import Stubber

from runway.cfngin.lookups.handlers.ami import AmiLookup, ImageNotFound

from ...factories import SessionStub, mock_provider

REGION = "us-east-1"


class TestAMILookup(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.ami.ImageNotFound."""

    client = boto3.client("ec2", region_name=REGION,
                          # bypass the need to have these in the env
                          aws_access_key_id='testing',
                          aws_secret_access_key='testing')

    def setUp(self):
        """Run before tests."""
        self.stubber = Stubber(self.client)
        self.provider = mock_provider(region=REGION)

    @mock.patch("runway.cfngin.lookups.handlers.ami.get_session",
                return_value=SessionStub(client))
    def test_basic_lookup_single_image(self, _mock_client):
        """Test basic lookup single image."""
        image_id = "ami-fffccc111"
        self.stubber.add_response(
            "describe_images",
            {
                "Images": [
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-13T01:17:44.000Z",
                        "State": "available",
                        "ImageId": image_id,
                        "Name": "Fake Image 1",
                        "VirtualizationType": "hvm",
                    }
                ]
            }
        )

        with self.stubber:
            value = AmiLookup.handle(
                value=r'owners:self name_regex:Fake\sImage\s\d',
                provider=self.provider
            )
            self.assertEqual(value, image_id)

    @mock.patch("runway.cfngin.lookups.handlers.ami.get_session",
                return_value=SessionStub(client))
    def test_basic_lookup_with_region(self, _mock_client):
        """Test basic lookup with region."""
        image_id = "ami-fffccc111"
        self.stubber.add_response(
            "describe_images",
            {
                "Images": [
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-13T01:17:44.000Z",
                        "State": "available",
                        "ImageId": image_id,
                        "Name": "Fake Image 1",
                        "VirtualizationType": "hvm",
                    }
                ]
            }
        )

        with self.stubber:
            value = AmiLookup.handle(
                value=r'us-west-1@owners:self name_regex:Fake\sImage\s\d',
                provider=self.provider
            )
            self.assertEqual(value, image_id)

    @mock.patch("runway.cfngin.lookups.handlers.ami.get_session",
                return_value=SessionStub(client))
    def test_basic_lookup_multiple_images(self, _mock_client):
        """Test basic lookup multiple images."""
        image_id = "ami-fffccc111"
        self.stubber.add_response(
            "describe_images",
            {
                "Images": [
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-13T01:17:44.000Z",
                        "State": "available",
                        "ImageId": "ami-fffccc110",
                        "Name": "Fake Image 1",
                        "VirtualizationType": "hvm",
                    },
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-14T01:17:44.000Z",
                        "State": "available",
                        "ImageId": image_id,
                        "Name": "Fake Image 2",
                        "VirtualizationType": "hvm",
                    },
                    # include an ARI in the response to ensure it can be handled
                    # (these don't include a 'Name')
                    {
                        'OwnerId': '897883143566',
                        'Architecture': 'x86_64',
                        'CreationDate': '2011-06-24T20:34:25.000Z',
                        'State': 'available',
                        'ImageId': 'ari-e6bc478f',
                        'VirtualizationType': 'paravirtual'
                    }
                ]
            }
        )

        with self.stubber:
            value = AmiLookup.handle(
                value=r'owners:self name_regex:Fake\sImage\s\d',
                provider=self.provider
            )
            self.assertEqual(value, image_id)

    @mock.patch("runway.cfngin.lookups.handlers.ami.get_session",
                return_value=SessionStub(client))
    def test_basic_lookup_multiple_images_name_match(self, _mock_client):
        """Test basic lookup multiple images name match."""
        image_id = "ami-fffccc111"
        self.stubber.add_response(
            "describe_images",
            {
                "Images": [
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-13T01:17:44.000Z",
                        "State": "available",
                        "ImageId": "ami-fffccc110",
                        "Name": "Fa---ke Image 1",
                        "VirtualizationType": "hvm",
                    },
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-14T01:17:44.000Z",
                        "State": "available",
                        "ImageId": image_id,
                        "Name": "Fake Image 2",
                        "VirtualizationType": "hvm",
                    },
                ]
            }
        )

        with self.stubber:
            value = AmiLookup.handle(
                value=r'owners:self name_regex:Fake\sImage\s\d',
                provider=self.provider
            )
            self.assertEqual(value, image_id)

    @mock.patch("runway.cfngin.lookups.handlers.ami.get_session",
                return_value=SessionStub(client))
    def test_basic_lookup_no_matching_images(self, _mock_client):
        """Test basic lookup no matching images."""
        self.stubber.add_response(
            "describe_images",
            {
                "Images": []
            }
        )

        with self.stubber:
            with self.assertRaises(ImageNotFound):
                AmiLookup.handle(
                    value=r'owners:self name_regex:Fake\sImage\s\d',
                    provider=self.provider
                )

    @mock.patch("runway.cfngin.lookups.handlers.ami.get_session",
                return_value=SessionStub(client))
    def test_basic_lookup_no_matching_images_from_name(self, _mock_client):
        """Test basic lookup no matching images from name."""
        image_id = "ami-fffccc111"
        self.stubber.add_response(
            "describe_images",
            {
                "Images": [
                    {
                        "OwnerId": "897883143566",
                        "Architecture": "x86_64",
                        "CreationDate": "2011-02-13T01:17:44.000Z",
                        "State": "available",
                        "ImageId": image_id,
                        "Name": "Fake Image 1",
                        "VirtualizationType": "hvm",
                    }
                ]
            }
        )

        with self.stubber:
            with self.assertRaises(ImageNotFound):
                AmiLookup.handle(
                    value=r'owners:self name_regex:MyImage\s\d',
                    provider=self.provider
                )



"""CFNgin lookup registry."""
import logging
import warnings

from six import string_types

from runway.lookups.handlers import ssm
from runway.util import load_object_from_string

from ..exceptions import FailedVariableLookup, UnknownLookupType
from .handlers import ami, default, dynamodb, envvar
from .handlers import file as file_handler
from .handlers import hook_data, kms, output, rxref, split, ssmstore, xref

CFNGIN_LOOKUP_HANDLERS = {}


def register_lookup_handler(lookup_type, handler_or_path):
    """Register a lookup handler.

    Args:
        lookup_type (str): Name to register the handler under.
        handler_or_path (Union[Callable, str]): A function or a path to a
            handler.

    """
    handler = handler_or_path
    if isinstance(handler_or_path, string_types):
        handler = load_object_from_string(handler_or_path)
    CFNGIN_LOOKUP_HANDLERS[lookup_type] = handler
    if not isinstance(handler, type):
        # Hander is a not a new-style handler
        logger = logging.getLogger(__name__)
        logger.warning("Registering lookup `%s`: Please upgrade to use the "
                       "new style of Lookups.", lookup_type)
        warnings.warn(
            # For some reason, this does not show up...
            # Leaving it in anyway
            "Lookup `%s`: Please upgrade to use the new style of Lookups"
            "." % lookup_type,
            DeprecationWarning,
            stacklevel=2,
        )


def unregister_lookup_handler(lookup_type):
    """Unregister the specified lookup type.

    This is useful when testing various lookup types if you want to unregister
    the lookup type after the test runs.

    Args:
        lookup_type (str): Name of the lookup type to unregister.

    """
    CFNGIN_LOOKUP_HANDLERS.pop(lookup_type, None)


def resolve_lookups(variable, context, provider):
    """Resolve a set of lookups.

    Args:
        variable (:class:`runway.cfngin.variables.Variable`): The variable
            resolving it's lookups.
        context (:class:`runway.cfngin.context.Context`): Context instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Provider
            instance.

    Returns:
        Dict[str, Any]: Lookup -> resolved value

    """
    resolved_lookups = {}
    for lookup in variable.lookups:
        try:
            handler = CFNGIN_LOOKUP_HANDLERS[lookup.type]
        except KeyError:
            raise UnknownLookupType(lookup)
        try:
            resolved_lookups[lookup] = handler(
                value=lookup.input,
                context=context,
                provider=provider,
            )
        except Exception as err:
            raise FailedVariableLookup(variable.name, lookup, err)
    return resolved_lookups


register_lookup_handler(ami.TYPE_NAME, ami.AmiLookup)
register_lookup_handler(default.TYPE_NAME, default.DefaultLookup)
register_lookup_handler(dynamodb.TYPE_NAME, dynamodb.DynamodbLookup)
register_lookup_handler(envvar.TYPE_NAME, envvar.EnvvarLookup)
register_lookup_handler(file_handler.TYPE_NAME, file_handler.FileLookup)
register_lookup_handler(hook_data.TYPE_NAME, hook_data.HookDataLookup)
register_lookup_handler(kms.TYPE_NAME, kms.KmsLookup)
register_lookup_handler(output.TYPE_NAME, output.OutputLookup)
register_lookup_handler(rxref.TYPE_NAME, rxref.RxrefLookup)
register_lookup_handler(split.TYPE_NAME, split.SplitLookup)
register_lookup_handler(ssm.TYPE_NAME, ssm.SsmLookup)
register_lookup_handler(ssmstore.TYPE_NAME, ssmstore.SsmstoreLookup)
register_lookup_handler(xref.TYPE_NAME, xref.XrefLookup)



"""User Pool Client Updater.

Responsible for updating the User Pool Client with the generated
distribution url + callback url paths.
"""

import logging
from typing import Any, Dict, Optional  # pylint: disable=unused-import

from runway.cfngin.providers.base import BaseProvider  # pylint: disable=unused-import
from runway.cfngin.context import Context  # noqa pylint: disable=unused-import
from runway.cfngin.session_cache import get_session

LOGGER = logging.getLogger(__name__)


def update(context,  # pylint: disable=unused-argument
           provider,
           **kwargs
          ):  # noqa: E124
    # type: (Context, BaseProvider, Optional[Dict[str, Any]]) -> bool
    """Update the callback urls for the User Pool Client.

    Required to match the redirect_uri being sent which contains
    our distribution and alternate domain names.

    Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    Keyword Args:
        alternate_domains (List[str]): A list of any alternate domains
            that need to be listed with the primary distribution domain
        redirect_path_sign_in (str): The redirect path after sign in
        redirect_path_sign_out (str): The redirect path after sign out
        oauth_scopes (List[str]): A list of all available validation
            scopes for oauth
    """
    session = get_session(provider.region)
    cognito_client = session.client('cognito-idp')

    # Combine alternate domains with main distribution
    redirect_domains = kwargs['alternate_domains'] + ['https://' + kwargs['distribution_domain']]

    # Create a list of all domains with their redirect paths
    redirect_uris_sign_in = [
        "%s%s" % (domain, kwargs['redirect_path_sign_in'])
        for domain in redirect_domains
    ]
    redirect_uris_sign_out = [
        "%s%s" % (domain, kwargs['redirect_path_sign_out'])
        for domain in redirect_domains
    ]
    # Update the user pool client
    try:
        cognito_client.update_user_pool_client(
            AllowedOAuthScopes=kwargs['oauth_scopes'],
            AllowedOAuthFlows=['code'],
            SupportedIdentityProviders=kwargs['supported_identity_providers'],
            AllowedOAuthFlowsUserPoolClient=True,
            ClientId=kwargs['client_id'],
            CallbackURLs=redirect_uris_sign_in,
            LogoutURLs=redirect_uris_sign_out,
            UserPoolId=context.hook_data['aae_user_pool_id_retriever']['id'],
        )
        return True
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.error('Was not able to update the callback urls on '
                     'the user pool client')
        LOGGER.error(err)
        return False



"""File lookup."""
# pylint: disable=arguments-differ,unused-argument
import base64
import json
import re

import yaml
from six import string_types
from six.moves.collections_abc import Mapping, Sequence  # pylint: disable=E
from troposphere import Base64, GenericHelperFn

from runway.lookups.handlers.base import LookupHandler

from ...util import read_value_from_path

TYPE_NAME = "file"

_PARAMETER_PATTERN = re.compile(r'{{([::|\w]+)}}')


class FileLookup(LookupHandler):
    """File lookup."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        r"""Translate a filename into the file contents.

        Args:
            value (str): Parameter(s) given to this lookup.
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Fields should use the following format::

            <codec>:<path>

        Example::

            # We've written a file to /some/path:
            $ echo "hello there" > /some/path

            # With CFNgin we would reference the contents of this file with the
            # following
            conf_key: ${file plain:file://some/path}

            # The above would resolve to
            conf_key: hello there

            # Or, if we used wanted a base64 encoded copy of the file data
            conf_key: ${file base64:file://some/path}

            # The above would resolve to
            conf_key: aGVsbG8gdGhlcmUK

        Supported codecs:

        **plain**
            Plain Text

        **base64**
            Encode the plain text file at the given path with base64 prior to
            returning it.

        **parameterized**
            The same as plain, but additionally supports referencing template
            parameters to create userdata that's supplemented with information
            from the template, as is commonly needed in EC2 UserData.
            For example, given a template parameter of BucketName, the file
            could contain the following text::

                #!/bin/sh
                aws s3 sync s3://{{BucketName}}/somepath /somepath

            Then you could use something like this in the YAML config
            file::

                UserData: ${file parameterized:/path/to/file}

            Resulting in the UserData parameter being defined as::

                { "Fn::Join" : ["", [
                    "#!/bin/sh\\naws s3 sync s3://",
                    {"Ref" : "BucketName"},
                    "/somepath /somepath"
                ]] }

        **parameterized-b64**
            The same as parameterized, with the results additionally wrapped
            in ``{ "Fn::Base64": ... }`` , which is what you actually need
            for EC2 UserData.

            When using parameterized-b64 for UserData, you should use a
            variable defined as such:

            .. code-block:: python

                from troposphere import AWSHelperFn

                "UserData": {
                    "type": AWSHelperFn,
                    "description": "Instance user data",
                    "default": Ref("AWS::NoValue")
                }

            Then assign UserData in a LaunchConfiguration or Instance to
            ``self.get_variables()["UserData"]``. Note that we use AWSHelperFn
            as the type because the parameterized-b64 codec returns either a
            Base64 or a GenericHelperFn troposphere object.

        **json**
            Decode the file as JSON and return the resulting object

        **json-parameterized**
            Same as ``json``, but applying templating rules from
            ``parameterized`` to every object *value*. Note that
            object *keys* are not modified. Example (an external
            PolicyDocument)::

                    {
                        "Version": "2012-10-17",
                        "Statement": [
                            {
                                "Effect": "Allow",
                                "Action": [
                                    "some:Action"
                                ],
                                "Resource": "{{MyResource}}"
                            }
                        ]
                    }

        **yaml**
            Decode the file as YAML and return the resulting object.
            All strings are returned as ``unicode`` even in Python 2.

        **yaml-parameterized**
            Same as ``json-parameterized``, but using YAML. Example::

                Version: 2012-10-17
                Statement:
                  - Effect: Allow
                    Action:
                      - "some:Action"
                    Resource: "{{MyResource}}"

        """
        try:
            codec, path = value.split(":", 1)
        except ValueError:
            raise TypeError(
                "File value must be of the format"
                " \"<codec>:<path>\" (got %s)" % (value)
            )

        value = read_value_from_path(path)

        return CODECS[codec](value)


def _parameterize_string(raw):
    """Substitute placeholders in a string using CloudFormation references.

    Args:
        raw (str): String to be processed. Byte strings are not
            supported; decode them before passing them to this function.

    Returns:
        str | :class:`troposphere.GenericHelperFn`: An expression with
            placeholders from the input replaced, suitable to be passed to
            Troposphere to be included in CloudFormation template. This will
            be the input string without modification if no substitutions are
            found, and a composition of CloudFormation calls otherwise.

    """
    parts = []
    s_index = 0

    for match in _PARAMETER_PATTERN.finditer(raw):
        parts.append(raw[s_index:match.start()])
        parts.append({u"Ref": match.group(1)})
        s_index = match.end()

    if not parts:
        return GenericHelperFn(raw)

    parts.append(raw[s_index:])
    return GenericHelperFn({u"Fn::Join": [u"", parts]})


def parameterized_codec(raw, b64):
    """Parameterize a string, possibly encoding it as Base64 afterwards.

    Args:
        raw (Union[bytes, str]): String to be processed. Byte strings will be
            interpreted as UTF-8.
        b64 (bool): Whether to wrap the output in a Base64 CloudFormation
            call.

    Returns:
        :class:`troposphere.AWSHelperFn`: Output to be included in a
        CloudFormation template.

    """
    if isinstance(raw, bytes):
        raw = raw.decode('utf-8')

    result = _parameterize_string(raw)

    # Note, since we want a raw JSON object (not a string) output in the
    # template, we wrap the result in GenericHelperFn (not needed if we're
    # using Base64)
    return Base64(result.data) if b64 else result


def _parameterize_obj(obj):
    """Recursively parameterize all strings contained in an object.

    Parametrizes all values of a Mapping, all items of a Sequence, an
    unicode string, or pass other objects through unmodified.

    Byte strings will be interpreted as UTF-8.

    Args:
        obj (Any): Data to parameterize.

    Return:
        A parameterized object to be included in a CloudFormation template.
        Mappings are converted to `dict`, Sequences are converted to  `list`,
        and strings possibly replaced by compositions of function calls.

    """
    if isinstance(obj, Mapping):
        return dict((key, _parameterize_obj(value))
                    for key, value in obj.items())
    if isinstance(obj, bytes):
        return _parameterize_string(obj.decode('utf8'))
    if isinstance(obj, string_types):
        return _parameterize_string(obj)
    if isinstance(obj, Sequence):
        return list(_parameterize_obj(item) for item in obj)
    return obj


class SafeUnicodeLoader(yaml.SafeLoader):
    """Safe unicode loader."""

    def construct_yaml_str(self, node):
        """Construct yaml str."""
        return self.construct_scalar(node)


def yaml_codec(raw, parameterized=False):
    """YAML codec."""
    data = yaml.load(raw, Loader=SafeUnicodeLoader)
    return _parameterize_obj(data) if parameterized else data


def json_codec(raw, parameterized=False):
    """JSON codec."""
    data = json.loads(raw)
    return _parameterize_obj(data) if parameterized else data


CODECS = {
    "plain": lambda x: x,
    "base64": lambda x: base64.b64encode(x.encode('utf8')).decode('utf-8'),
    "parameterized": lambda x: parameterized_codec(x, False),
    "parameterized-b64": lambda x: parameterized_codec(x, True),
    "yaml": lambda x: yaml_codec(x, parameterized=False),
    "yaml-parameterized": lambda x: yaml_codec(x, parameterized=True),
    "json": lambda x: json_codec(x, parameterized=False),
    "json-parameterized": lambda x: json_codec(x, parameterized=True),
}



"""Tests for runway.cfngin.lookups.handlers.xref."""
import unittest

from mock import MagicMock

from runway.cfngin.lookups.handlers.xref import XrefLookup


class TestXrefHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.xref.XrefHandler."""

    def setUp(self):
        """Run before tests."""
        self.provider = MagicMock()
        self.context = MagicMock()

    def test_xref_handler(self):
        """Test xref handler."""
        self.provider.get_output.return_value = "Test Output"
        value = XrefLookup.handle("fully-qualified-stack-name::SomeOutput",
                                  provider=self.provider,
                                  context=self.context)
        self.assertEqual(value, "Test Output")
        self.assertEqual(self.context.get_fqn.call_count, 0)
        args = self.provider.get_output.call_args
        self.assertEqual(args[0][0], "fully-qualified-stack-name")
        self.assertEqual(args[0][1], "SomeOutput")



"""Tests for runway.cfngin.lookups.handlers.split."""
import unittest

from runway.cfngin.lookups.handlers.split import SplitLookup


class TestSplitLookup(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.split.SplitLookup."""

    def test_single_character_split(self):
        """Test single character split."""
        value = ",::a,b,c"
        expected = ["a", "b", "c"]
        assert SplitLookup.handle(value) == expected

    def test_multi_character_split(self):
        """Test multi character split."""
        value = ",,::a,,b,c"
        expected = ["a", "b,c"]
        assert SplitLookup.handle(value) == expected

    def test_invalid_value_split(self):
        """Test invalid value split."""
        value = ",:a,b,c"
        with self.assertRaises(ValueError):
            SplitLookup.handle(value)



"""AWS Route 53 hook."""
# pylint: disable=unused-argument
import logging

from ..session_cache import get_session
from ..util import create_route53_zone

LOGGER = logging.getLogger(__name__)


def create_domain(provider, context, **kwargs):
    """Create a domain within route53.

    Args:
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Provider
            instance. (passed in by CFNgin)
        context (:class:`runway.cfngin.context.Context`): Context instance.
            (passed in by CFNgin)

    Keyword Args:
        domain (str): Domain name for the Route 53 hosted zone to be
            created.

    Returns:
        Dict[str, str]: Dict containing ``domain`` and ``zone_id``.

    """
    session = get_session(provider.region)
    client = session.client("route53")
    domain = kwargs.get("domain")
    if not domain:
        LOGGER.error("domain argument or BaseDomain variable not provided.")
        return False
    zone_id = create_route53_zone(client, domain)
    return {"domain": domain, "zone_id": zone_id}



"""Tests for runway.cfngin.actions.build."""
# pylint: disable=no-self-use,protected-access,unused-argument
import unittest
from collections import namedtuple

from mock import MagicMock, PropertyMock, patch

from runway.cfngin import exceptions
from runway.cfngin.actions import build
from runway.cfngin.actions.build import (UsePreviousParameterValue,
                                         _handle_missing_parameters,
                                         _resolve_parameters)
from runway.cfngin.blueprints.variables.types import CFNString
from runway.cfngin.context import Config, Context
from runway.cfngin.exceptions import StackDidNotChange, StackDoesNotExist
from runway.cfngin.plan import Graph, Plan, Step
from runway.cfngin.providers.aws.default import Provider
from runway.cfngin.providers.base import BaseProvider
from runway.cfngin.session_cache import get_session
from runway.cfngin.status import (COMPLETE, FAILED, PENDING, SKIPPED,
                                  SUBMITTED, NotSubmittedStatus)

from ..factories import MockProviderBuilder, MockThreadingEvent


def mock_stack_parameters(parameters):
    """Mock stack parameters."""
    return {
        'Parameters': [
            {'ParameterKey': k, 'ParameterValue': v}
            for k, v in parameters.items()
        ]
    }


class MockProvider(BaseProvider):
    """Mock provider."""

    def __init__(self, *args, **kwargs):
        """Instantiate class."""
        self._outputs = kwargs.get('outputs', {})

    def set_outputs(self, outputs):
        """Set outputs."""
        self._outputs = outputs

    def get_stack(self, stack_name, *args, **kwargs):
        """Get stack."""
        if stack_name not in self._outputs:
            raise exceptions.StackDoesNotExist(stack_name)
        return {"name": stack_name, "outputs": self._outputs[stack_name]}

    def get_outputs(self, stack_name, *args, **kwargs):
        """Get outputs."""
        stack = self.get_stack(stack_name)
        return stack["outputs"]


class TestBuildAction(unittest.TestCase):
    """Tests for runway.cfngin.actions.build.BuildAction."""

    def setUp(self):
        """Run before tests."""
        self.context = Context(config=Config({"namespace": "namespace"}))
        self.provider = MockProvider()
        self.build_action = build.Action(
            self.context,
            provider_builder=MockProviderBuilder(self.provider))

    def _get_context(self, extra_config_args=None, **kwargs):
        """Get context."""
        config = {
            "namespace": "namespace",
            "stacks": [
                {"name": "vpc"},
                {
                    "name": "bastion",
                    "variables": {
                        "test": "${output vpc::something}"
                    }
                },
                {
                    "name": "db",
                    "variables": {
                        "test": "${output vpc::something}",
                        "else": "${output bastion::something}"
                    }
                },
                {"name": "other", "variables": {}}
            ],
        }
        if extra_config_args:
            config.update(extra_config_args)
        return Context(config=Config(config), **kwargs)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    def test_generate_plan_persist_destroy(self, mock_graph_tags):
        """Test generate plan persist destroy."""
        mock_graph_tags.return_value = {}
        context = self._get_context(
            extra_config_args={'persistent_graph_key': 'test.json'}
        )
        context._persistent_graph = Graph.from_steps(
            [Step.from_stack_name('removed', context)]
        )
        build_action = build.Action(context=context)
        plan = build_action._Action__generate_plan()

        self.assertIsInstance(plan, Plan)
        self.assertEqual(build.Action.DESCRIPTION, plan.description)
        mock_graph_tags.assert_called_once()
        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(5, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict['other'])
        self.assertEqual(set(), result_graph_dict['removed'])
        self.assertEqual(set(), result_graph_dict['vpc'])
        self.assertEqual(set(['vpc']), result_graph_dict['bastion'])
        self.assertEqual(set(['bastion', 'vpc']), result_graph_dict['db'])
        self.assertEqual(build_action._destroy_stack,
                         plan.graph.steps['removed'].fn)
        self.assertEqual(build_action._launch_stack,
                         plan.graph.steps['vpc'].fn)
        self.assertEqual(build_action._launch_stack,
                         plan.graph.steps['bastion'].fn)
        self.assertEqual(build_action._launch_stack,
                         plan.graph.steps['db'].fn)
        self.assertEqual(build_action._launch_stack,
                         plan.graph.steps['other'].fn)

    def test_handle_missing_params(self):
        """Test handle missing params."""
        existing_stack_param_dict = {
            "StackName": "teststack",
            "Address": "192.168.0.1"
        }
        existing_stack_params = mock_stack_parameters(
            existing_stack_param_dict
        )
        all_params = existing_stack_param_dict.keys()
        required = ["Address"]
        parameter_values = {"Address": "192.168.0.1"}
        expected_params = {"StackName": UsePreviousParameterValue,
                           "Address": "192.168.0.1"}
        result = _handle_missing_parameters(parameter_values, all_params,
                                            required, existing_stack_params)
        self.assertEqual(sorted(result), sorted(list(expected_params.items())))

    def test_missing_params_no_existing_stack(self):
        """Test missing params no existing stack."""
        all_params = ["Address", "StackName"]
        required = ["Address"]
        parameter_values = {}
        with self.assertRaises(exceptions.MissingParameterException) as result:
            _handle_missing_parameters(parameter_values, all_params, required)

        self.assertEqual(result.exception.parameters, required)

    def test_existing_stack_params_does_not_override_given_params(self):
        """Test existing stack params does not override given params."""
        existing_stack_param_dict = {
            "StackName": "teststack",
            "Address": "192.168.0.1"
        }
        existing_stack_params = mock_stack_parameters(
            existing_stack_param_dict
        )
        all_params = existing_stack_param_dict.keys()
        required = ["Address"]
        parameter_values = {"Address": "10.0.0.1"}
        result = _handle_missing_parameters(parameter_values, all_params,
                                            required, existing_stack_params)
        self.assertEqual(
            sorted(result),
            sorted(list(parameter_values.items()))
        )

    def test_generate_plan(self):
        """Test generate plan."""
        context = self._get_context()
        build_action = build.Action(context, cancel=MockThreadingEvent())
        plan = build_action._Action__generate_plan()
        self.assertEqual(
            {
                'db': set(['bastion', 'vpc']),
                'bastion': set(['vpc']),
                'other': set([]),
                'vpc': set([])},
            plan.graph.to_dict()
        )

    def test_does_not_execute_plan_when_outline_specified(self):
        """Test does not execute plan when outline specified."""
        context = self._get_context()
        build_action = build.Action(context, cancel=MockThreadingEvent())
        with patch.object(build_action, "_generate_plan") as \
                mock_generate_plan:
            build_action.run(outline=True)
            self.assertEqual(mock_generate_plan().execute.call_count, 0)

    def test_execute_plan_when_outline_not_specified(self):
        """Test execute plan when outline not specified."""
        context = self._get_context()
        build_action = build.Action(context, cancel=MockThreadingEvent())
        with patch.object(build_action, "_generate_plan") as \
                mock_generate_plan:
            build_action.run(outline=False)
            self.assertEqual(mock_generate_plan().execute.call_count, 1)

    @patch('runway.cfngin.context.Context._persistent_graph_tags',
           new_callable=PropertyMock)
    @patch('runway.cfngin.context.Context.lock_persistent_graph',
           new_callable=MagicMock)
    @patch('runway.cfngin.context.Context.unlock_persistent_graph',
           new_callable=MagicMock)
    @patch('runway.cfngin.plan.Plan.execute', new_callable=MagicMock)
    def test_run_persist(self, mock_execute, mock_unlock, mock_lock,
                         mock_graph_tags):
        """Test run persist."""
        mock_graph_tags.return_value = {}
        context = self._get_context(
            extra_config_args={'persistent_graph_key': 'test.json'}
        )
        context._persistent_graph = Graph.from_steps(
            [Step.from_stack_name('removed', context)]
        )
        build_action = build.Action(context=context)
        build_action.run()

        mock_graph_tags.assert_called_once()
        mock_lock.assert_called_once()
        mock_execute.assert_called_once()
        mock_unlock.assert_called_once()

    def test_should_update(self):
        """Test should update."""
        test_scenario = namedtuple("test_scenario",
                                   ["locked", "force", "result"])
        test_scenarios = (
            test_scenario(locked=False, force=False, result=True),
            test_scenario(locked=False, force=True, result=True),
            test_scenario(locked=True, force=False, result=False),
            test_scenario(locked=True, force=True, result=True)
        )
        mock_stack = MagicMock(["locked", "force", "name"])
        mock_stack.name = "test-stack"
        for test in test_scenarios:
            mock_stack.locked = test.locked
            mock_stack.force = test.force
            self.assertEqual(build.should_update(mock_stack), test.result)

    def test_should_ensure_cfn_bucket(self):
        """Test should ensure cfn bucket."""
        test_scenarios = [
            {"outline": False, "dump": False, "result": True},
            {"outline": True, "dump": False, "result": False},
            {"outline": False, "dump": True, "result": False},
            {"outline": True, "dump": True, "result": False},
            {"outline": True, "dump": "DUMP", "result": False}
        ]

        for scenario in test_scenarios:
            outline = scenario["outline"]
            dump = scenario["dump"]
            result = scenario["result"]
            try:
                self.assertEqual(
                    build.should_ensure_cfn_bucket(outline, dump), result)
            except AssertionError as err:
                err.args += ("scenario", str(scenario))
                raise

    def test_should_submit(self):
        """Test should submit."""
        test_scenario = namedtuple("test_scenario",
                                   ["enabled", "result"])
        test_scenarios = (
            test_scenario(enabled=False, result=False),
            test_scenario(enabled=True, result=True),
        )

        mock_stack = MagicMock(["enabled", "name"])
        mock_stack.name = "test-stack"
        for test in test_scenarios:
            mock_stack.enabled = test.enabled
            self.assertEqual(build.should_submit(mock_stack), test.result)


class TestLaunchStack(TestBuildAction):
    """Tests for runway.cfngin.actions.build.BuildAction launch stack."""

    def setUp(self):
        """Run before tests."""
        self.context = self._get_context()
        self.session = get_session(region=None)
        self.provider = Provider(self.session, interactive=False,
                                 recreate_failed=False)
        provider_builder = MockProviderBuilder(self.provider)
        self.build_action = build.Action(self.context,
                                         provider_builder=provider_builder,
                                         cancel=MockThreadingEvent())

        self.stack = MagicMock()
        self.stack.region = None
        self.stack.name = 'vpc'
        self.stack.fqn = 'vpc'
        self.stack.blueprint.rendered = '{}'
        self.stack.locked = False
        self.stack_status = None

        plan = self.build_action._Action__generate_plan()
        self.step = plan.steps[0]
        self.step.stack = self.stack

        def patch_object(*args, **kwargs):
            mock_object = patch.object(*args, **kwargs)
            self.addCleanup(mock_object.stop)
            mock_object.start()

        def get_stack(name, *args, **kwargs):
            if name != self.stack.name or not self.stack_status:
                raise StackDoesNotExist(name)

            return {'StackName': self.stack.name,
                    'StackStatus': self.stack_status,
                    'Outputs': [],
                    'Tags': []}

        def get_events(name, *args, **kwargs):
            return [{'ResourceStatus': 'ROLLBACK_IN_PROGRESS',
                     'ResourceStatusReason': 'CFN fail'}]

        patch_object(self.provider, 'get_stack', side_effect=get_stack)
        patch_object(self.provider, 'update_stack')
        patch_object(self.provider, 'create_stack')
        patch_object(self.provider, 'destroy_stack')
        patch_object(self.provider, 'get_events', side_effect=get_events)

        patch_object(self.build_action, "s3_stack_push")

    def _advance(self, new_provider_status, expected_status, expected_reason):
        """Advance."""
        self.stack_status = new_provider_status
        status = self.step._run_once()
        self.assertEqual(status, expected_status)
        self.assertEqual(status.reason, expected_reason)

    def test_launch_stack_disabled(self):
        """Test launch stack disabled."""
        self.assertEqual(self.step.status, PENDING)

        self.stack.enabled = False
        self._advance(None, NotSubmittedStatus(), "disabled")

    def test_launch_stack_create(self):
        """Test launch stack create."""
        # initial status should be PENDING
        self.assertEqual(self.step.status, PENDING)

        # initial run should return SUBMITTED since we've passed off to CF
        self._advance(None, SUBMITTED, "creating new stack")

        # status should stay as SUBMITTED when the stack becomes available
        self._advance('CREATE_IN_PROGRESS', SUBMITTED, "creating new stack")

        # status should become COMPLETE once the stack finishes
        self._advance('CREATE_COMPLETE', COMPLETE, "creating new stack")

    def test_launch_stack_create_rollback(self):
        """Test launch stack create rollback."""
        # initial status should be PENDING
        self.assertEqual(self.step.status, PENDING)

        # initial run should return SUBMITTED since we've passed off to CF
        self._advance(None, SUBMITTED, "creating new stack")

        # provider should now return the CF stack since it exists
        self._advance("CREATE_IN_PROGRESS", SUBMITTED,
                      "creating new stack")

        # rollback should be noticed
        self._advance("ROLLBACK_IN_PROGRESS", SUBMITTED,
                      "rolling back new stack")

        # rollback should not be added twice to the reason
        self._advance("ROLLBACK_IN_PROGRESS", SUBMITTED,
                      "rolling back new stack")

        # rollback should finish with failure
        self._advance("ROLLBACK_COMPLETE", FAILED,
                      "rolled back new stack")

    def test_launch_stack_recreate(self):
        """Test launch stack recreate."""
        self.provider.recreate_failed = True  # pylint: disable=attribute-defined-outside-init

        # initial status should be PENDING
        self.assertEqual(self.step.status, PENDING)

        # first action with an existing failed stack should be deleting it
        self._advance("ROLLBACK_COMPLETE", SUBMITTED,
                      "destroying stack for re-creation")

        # status should stay as submitted during deletion
        self._advance("DELETE_IN_PROGRESS", SUBMITTED,
                      "destroying stack for re-creation")

        # deletion being complete must trigger re-creation
        self._advance("DELETE_COMPLETE", SUBMITTED,
                      "re-creating stack")

        # re-creation should continue as SUBMITTED
        self._advance("CREATE_IN_PROGRESS", SUBMITTED,
                      "re-creating stack")

        # re-creation should finish with success
        self._advance("CREATE_COMPLETE", COMPLETE,
                      "re-creating stack")

    def test_launch_stack_update_skipped(self):
        """Test launch stack update skipped."""
        # initial status should be PENDING
        self.assertEqual(self.step.status, PENDING)

        # start the upgrade, that will be skipped
        self.provider.update_stack.side_effect = StackDidNotChange
        self._advance("CREATE_COMPLETE", SKIPPED,
                      "nochange")

    def test_launch_stack_update_rollback(self):
        """Test launch stack update rollback."""
        # initial status should be PENDING
        self.assertEqual(self.step.status, PENDING)

        # initial run should return SUBMITTED since we've passed off to CF
        self._advance("CREATE_COMPLETE", SUBMITTED,
                      "updating existing stack")

        # update should continue as SUBMITTED
        self._advance("UPDATE_IN_PROGRESS", SUBMITTED,
                      "updating existing stack")

        # rollback should be noticed
        self._advance("UPDATE_ROLLBACK_IN_PROGRESS", SUBMITTED,
                      "rolling back update")

        # rollback should finish with failure
        self._advance("UPDATE_ROLLBACK_COMPLETE", FAILED,
                      "rolled back update")

    def test_launch_stack_update_success(self):
        """Test launch stack update success."""
        # initial status should be PENDING
        self.assertEqual(self.step.status, PENDING)

        # initial run should return SUBMITTED since we've passed off to CF
        self._advance("CREATE_COMPLETE", SUBMITTED,
                      "updating existing stack")

        # update should continue as SUBMITTED
        self._advance("UPDATE_IN_PROGRESS", SUBMITTED,
                      "updating existing stack")

        # update should finish with success
        self._advance("UPDATE_COMPLETE", COMPLETE,
                      "updating existing stack")


class TestFunctions(unittest.TestCase):
    """Tests for runway.cfngin.actions.build module level functions."""

    def setUp(self):
        """Run before tests."""
        self.ctx = Context({"namespace": "test"})
        self.prov = MagicMock()
        self.blueprint = MagicMock()

    def test_resolve_parameters_unused_parameter(self):
        """Test resolve parameters unused parameter."""
        self.blueprint.get_parameter_definitions.return_value = {
            "a": {
                "type": CFNString,
                "description": "A"},
            "b": {
                "type": CFNString,
                "description": "B"}
        }
        params = {"a": "Apple", "c": "Carrot"}
        resolved_params = _resolve_parameters(params, self.blueprint)
        self.assertNotIn("c", resolved_params)
        self.assertIn("a", resolved_params)

    def test_resolve_parameters_none_conversion(self):
        """Test resolve parameters none conversion."""
        self.blueprint.get_parameter_definitions.return_value = {
            "a": {
                "type": CFNString,
                "description": "A"},
            "b": {
                "type": CFNString,
                "description": "B"}
        }
        params = {"a": None, "c": "Carrot"}
        resolved_params = _resolve_parameters(params, self.blueprint)
        self.assertNotIn("a", resolved_params)

    def test_resolve_parameters_booleans(self):
        """Test resolve parameters booleans."""
        self.blueprint.get_parameter_definitions.return_value = {
            "a": {
                "type": CFNString,
                "description": "A"},
            "b": {
                "type": CFNString,
                "description": "B"},
        }
        params = {"a": True, "b": False}
        resolved_params = _resolve_parameters(params, self.blueprint)
        self.assertEqual("true", resolved_params["a"])
        self.assertEqual("false", resolved_params["b"])



"""
Abstract parent class for a 'Source' type object.

Allows us to specify specific remote sourced resources for out application
(Git, S3, ect.)
"""
# pylint: disable=unused-import
from typing import Dict, Optional, Union  # noqa: F401

import os
import logging

LOGGER = logging.getLogger('runway')


class Source(object):
    """Abstract parent class for a 'Source' type object.

    The Source parent class allows us to specify remote resources
    for our application via services such as Git or S3.  A cache
    directory, as part of object's configuration, is automatically
    created by default in the users home directory: ``~/.runway_cache``.
    This folder can be overridden by specifying the ``cache_dir`` property
    in the configuration passed.

    Every Source type object is expected to have a ``fetch`` method which
    will return the folder path at where the module requested resides.
    """

    def __init__(self, cache_dir='', **_):
        # type(Dict[str, Union[str, Dict[str, str]]]) -> Source
        """Source.

        Keyword Arguments:
            cache_dir (str): The directory where the given remote resource
                should be cached

        """
        self.cache_dir = cache_dir

        if not self.cache_dir:
            self.cache_dir = os.path.expanduser("~/.runway_cache")  # type: str

        self.__create_cache_directory()

    def fetch(self):
        # type: () -> None
        """Retrieve remote source. To be implemented in each subclass."""
        raise NotImplementedError

    def __create_cache_directory(self):
        # type: () -> None
        """If no cache directory exists for the remote runway modules, create one."""
        if not os.path.isdir(self.cache_dir):
            os.mkdir(self.cache_dir)

    @staticmethod
    def sanitize_directory_path(uri):
        # type: (str) -> str
        """Sanitize a Source directory path string.

        Keyword Arguments:
            uri (str): The uniform resource identifier when targetting a remote resource.

        """
        for i in ['@', '/', ':']:
            uri = uri.replace(i, '_')  # type: str
        return uri



"""CFNgin destroy action."""
import logging

from ..exceptions import StackDoesNotExist
from ..hooks.utils import handle_hooks
from ..status import INTERRUPTED, PENDING, SUBMITTED, CompleteStatus
from ..status import StackDoesNotExist as StackDoesNotExistStatus
from ..status import SubmittedStatus
from .base import STACK_POLL_TIME, BaseAction, build_walker

LOGGER = logging.getLogger(__name__)

DESTROYED_STATUS = CompleteStatus("stack destroyed")
DESTROYING_STATUS = SubmittedStatus("submitted for destruction")


class Action(BaseAction):
    """Responsible for destroying CloudFormation stacks.

    Generates a destruction plan based on stack dependencies. Stack
    dependencies are reversed from the build action. For example, if a Stack B
    requires Stack A during build, during destroy Stack A requires Stack B be
    destroyed first.

    The plan defaults to printing an outline of what will be destroyed. If
    forced to execute, each stack will get destroyed in order.

    """

    DESCRIPTION = 'Destroy stacks'

    @property
    def _stack_action(self):
        """Run against a step."""
        return self._destroy_stack

    def _destroy_stack(self, stack, **kwargs):
        old_status = kwargs.get("status")
        wait_time = 0 if old_status is PENDING else STACK_POLL_TIME
        if self.cancel.wait(wait_time):
            return INTERRUPTED

        provider = self.build_provider(stack)

        try:
            provider_stack = provider.get_stack(stack.fqn)
        except StackDoesNotExist:
            LOGGER.debug("Stack %s does not exist.", stack.fqn)
            # Once the stack has been destroyed, it doesn't exist. If the
            # status of the step was SUBMITTED, we know we just deleted it,
            # otherwise it should be skipped
            if kwargs.get("status", None) == SUBMITTED:
                return DESTROYED_STATUS
            return StackDoesNotExistStatus()

        LOGGER.debug(
            "Stack %s provider status: %s",
            provider.get_stack_name(provider_stack),
            provider.get_stack_status(provider_stack),
        )
        if provider.is_stack_destroyed(provider_stack):
            return DESTROYED_STATUS
        if provider.is_stack_in_progress(provider_stack):
            return DESTROYING_STATUS
        LOGGER.debug("Destroying stack: %s", stack.fqn)
        provider.destroy_stack(provider_stack)
        return DESTROYING_STATUS

    def pre_run(self, **kwargs):
        """Any steps that need to be taken prior to running the action."""
        pre_destroy = self.context.config.pre_destroy
        if not kwargs.get('outline') and pre_destroy:
            handle_hooks(
                stage="pre_destroy",
                hooks=pre_destroy,
                provider=self.provider,
                context=self.context)

    def run(self, **kwargs):
        """Kicks off the destruction of the stacks in the stack_definitions."""
        plan = self._generate_plan(tail=kwargs.get('tail'), reverse=True,
                                   include_persistent_graph=True)
        if not plan.keys():
            LOGGER.warning('WARNING: No stacks detected (error in config?)')
        if kwargs.get('force', False):
            # need to generate a new plan to log since the outline sets the
            # steps to COMPLETE in order to log them
            plan.outline(logging.DEBUG)
            self.context.lock_persistent_graph(plan.lock_code)
            walker = build_walker(kwargs.get('concurrency', 0))
            try:
                plan.execute(walker)
            finally:
                self.context.unlock_persistent_graph(plan.lock_code)
        else:
            plan.outline(message="To execute this plan, run with "
                                 "\"--force\" flag.")

    def post_run(self, **kwargs):
        """Any steps that need to be taken after running the action."""
        post_destroy = self.context.config.post_destroy
        if not kwargs.get('outline') and post_destroy:
            handle_hooks(
                stage="post_destroy",
                hooks=post_destroy,
                provider=self.provider,
                context=self.context)



"""Tests for lookup registry and common lookup functionality."""
# pylint: disable=no-self-use
from runway.lookups.registry import RUNWAY_LOOKUP_HANDLERS
from runway.util import MutableMap

VALUES = {
    'str_val': 'test'
}
CONTEXT = MutableMap(**{
    'env_vars': VALUES
})
VARIABLES = MutableMap(**VALUES)


class TestCommonLookupFunctionality(object):
    """Test common lookup functionally.

    All lookup handles should be able to pass these tests. Handling must
    be implement in all lookups in the "handle" method using the provided
    class methods of the "LookupHandler" base class.

    """

    def test_handle_default(self, runway_context):
        """Verify default value is handled by lookups."""
        lookup_handlers = RUNWAY_LOOKUP_HANDLERS.copy()
        lookup_handlers.pop('ssm')  # requires special testing
        for _, lookup in lookup_handlers.items():
            query = 'NOT_VALID::default=default value'
            result = lookup.handle(query, context=runway_context,
                                   variables=VARIABLES)

            assert result == 'default value'

    def test_handle_transform(self, runway_context):
        """Verify transform is handled by lookup."""
        lookup_handlers = RUNWAY_LOOKUP_HANDLERS.copy()
        lookup_handlers.pop('ssm')  # requires special testing
        runway_context.env_vars.update(VALUES)

        for _, lookup in lookup_handlers.items():
            query = 'NOT_VALID::default=false, transform=bool'
            result = lookup.handle(query, context=runway_context,
                                   variables=VARIABLES)

            assert not result



"""Tests for runway.cfngin.actions.diff."""
import unittest

from operator import attrgetter
from runway.cfngin.actions.diff import (
    diff_dictionaries,
    diff_parameters,
    DictValue
)


class TestDictValueFormat(unittest.TestCase):
    """Tests for runway.cfngin.actions.diff.DictValue."""

    def test_status(self):
        """Test status."""
        added = DictValue("k0", None, "value_0")
        self.assertEqual(added.status(), DictValue.ADDED)
        removed = DictValue("k1", "value_1", None)
        self.assertEqual(removed.status(), DictValue.REMOVED)
        modified = DictValue("k2", "value_1", "value_2")
        self.assertEqual(modified.status(), DictValue.MODIFIED)
        unmodified = DictValue("k3", "value_1", "value_1")
        self.assertEqual(unmodified.status(), DictValue.UNMODIFIED)

    def test_format(self):
        """Test format."""
        added = DictValue("k0", None, "value_0")
        self.assertEqual(added.changes(),
                         ['+%s = %s' % (added.key, added.new_value)])
        removed = DictValue("k1", "value_1", None)
        self.assertEqual(removed.changes(),
                         ['-%s = %s' % (removed.key, removed.old_value)])
        modified = DictValue("k2", "value_1", "value_2")
        self.assertEqual(modified.changes(), [
            '-%s = %s' % (modified.key, modified.old_value),
            '+%s = %s' % (modified.key, modified.new_value)
        ])
        unmodified = DictValue("k3", "value_1", "value_1")
        self.assertEqual(unmodified.changes(), [' %s = %s' % (
            unmodified.key, unmodified.old_value)])
        self.assertEqual(unmodified.changes(), [' %s = %s' % (
            unmodified.key, unmodified.new_value)])


class TestDiffDictionary(unittest.TestCase):
    """Tests for runway.cfngin.actions.diff.diff_dictionaries."""

    def test_diff_dictionaries(self):
        """Test diff dictionaries."""
        old_dict = {
            "a": "Apple",
            "b": "Banana",
            "c": "Corn",
        }
        new_dict = {
            "a": "Apple",
            "b": "Bob",
            "d": "Doug",
        }

        [count, changes] = diff_dictionaries(old_dict, new_dict)
        self.assertEqual(count, 3)
        expected_output = [
            DictValue("a", "Apple", "Apple"),
            DictValue("b", "Banana", "Bob"),
            DictValue("c", "Corn", None),
            DictValue("d", None, "Doug"),
        ]
        expected_output.sort(key=attrgetter("key"))

        # compare all the outputs to the expected change
        for expected_change in expected_output:
            change = changes.pop(0)
            self.assertEqual(change, expected_change)

        # No extra output
        self.assertEqual(len(changes), 0)


class TestDiffParameters(unittest.TestCase):
    """Tests for runway.cfngin.actions.diff.diff_parameters."""

    def test_diff_parameters_no_changes(self):
        """Test diff parameters no changes."""
        old_params = {
            "a": "Apple"
        }
        new_params = {
            "a": "Apple"
        }

        param_diffs = diff_parameters(old_params, new_params)
        self.assertEqual(param_diffs, [])



"""Tests for runway commands."""
import os

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import (import_tests, execute_tests)


class Commands(IntegrationTest):
    """Base class for all runway command testing testing."""

    base_dir = os.path.abspath(os.path.dirname(__file__))
    tests_dir = os.path.join(base_dir, 'tests')

    def run(self):
        """Find all tests and run them."""
        suffix = os.getenv('COMMAND_SUFFIX', '*')
        pattern = 'test_{0}'.format(suffix)
        self.set_env_var('AWS_DEFAULT_REGION', 'us-east-1')
        import_tests(self.logger, self.tests_dir, pattern)
        tests = [test(self.logger, self.environment)
                 for test in Commands.__subclasses__()]
        if not tests:
            raise Exception('No tests were found.')
        self.logger.debug('FOUND TESTS: %s', tests)
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0  # assert that all subtests were successful
        return err_count

    def teardown(self):
        """Teardown resources create during init."""
        self.logger.debug('Nothing to do during test teardown')



"""CFNgin test."""
# flake8: noqa
# pylint: disable=invalid-name
from os.path import basename

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestLockedStack(Cfngin):
    """Test CFNgin with a locked stack.

    Requires valid AWS credentials.

    """

    REQUIRED_FIXTURE_FILES = [
        FILE_BASENAME + '.1.yaml',
        FILE_BASENAME + '.2.yaml'
    ]
    TEST_NAME = __name__

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        code, _stdout, stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            'Using default AWS provider mode',
            'locked-stack-vpc: submitted (creating new stack)',
            'locked-stack-vpc: complete (creating new stack)',
            'locked-stack-vpc: skipped (locked)',
            'locked-stack-bastion: submitted (creating new stack)',
            'locked-stack-bastion: complete (creating new stack)'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.runway_cmd('destroy')  # cleanup incase of failure
        self.cleanup_fixtures()
